local MAJOR = "StarLibWidgetHistogram-1.0" 
local MINOR = 13

assert(LibStub, MAJOR.." requires LibStub") 
local WidgetHistogram = LibStub:NewLibrary(MAJOR, MINOR)
if not WidgetHistogram then return end

local LibProperty = LibStub("StarLibProperty-1.0", true)
assert(LibProperty, MAJOR .. " requires StarLibProperty-1.0")
local LibTimer = LibStub("StarLibTimer-1.0", true)
assert(LibTimer, MAJOR .. " requires StarLibTimer-1.0")
local LibError = LibStub("StarLibError-1.0", true)
assert(LibError, MAJOR .. " requires StarLibError-1.0")
local LibWidget = LibStub("StarLibWidget-1.0", true)
assert(LibWidget, MAJOR .. " requires StarLibWidget-1.0")

local pool = setmetatable({}, {__mode = "k"})

if not WidgetHistogram.__index then
	WidgetHistogram.__index = WidgetHistogram
end

SCHARS = {
    {31, 31, 31, 31, 31, 31, 31, 31},
    {31, 16, 16, 16, 16, 16, 16, 31},
    {31,  0,  0,  0,  0,  0,  0, 31},
    {31,  1,  1,  1,  1,  1,  1, 31},
    {31, 31, 31, 31,  0,  0,  0,  0},
    { 0,  0,  0,  0, 31, 31, 31, 31}};
	
WidgetHistogram.DIR_EAST, WidgetHistogram.DIR_WEST = 1, 2

WidgetHistogram.directionList = {"East", "West"}
	
WidgetHistogram.defaults = {
	direction = WidgetHistogram.DIR_EAST, update = 500, width = 6, height = 50, char = "#"
}

local widgetType = {histogram=true, rc=true, special=true}

-- Create a new StarLibWidgetHistogram object
-- @name WidgetHistogram.New
-- @usage WidgetHistogram:New(visitor, name, config, row, col, layer, errorLevel, callback, data)
-- @param visitor An StarLibCore-1.0 object
-- @param config This widget's settings
-- @param row This widget's row
-- @param col This widget's column
-- @param layer This widget's layer
-- @param errorLevel This widget's verboseness level
-- @param callback This widget's draw function
-- @param data Data to pass back through the callback
-- @return A new StarLibWidgetHistogram object

function WidgetHistogram:New(visitor, name, config, row, col, layer, errorLevel, callback, data)

	assert(name, "Please provide the histogram with a name")
	assert(config, "Please provide the histogram with a config")
	assert(config.expression, name .. ": Please provide the histogram with an expression")
			
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
		obj.history = {}
	end

	setmetatable(obj, self)
	
	obj.error = LibError:New(MAJOR .. ": " .. name, errorLevel)
		
	obj.errorLevel = errorLevel

	obj.callback = callback
	obj.data = data

	obj.widget = LibWidget:New(obj, visitor, name, config, row, col, layer, widgetType, env, errorLevel)


    obj.callback = callback
	obj.data = data

    obj.min = 0
	obj.max = 0
    obj.string = ""

    obj.expression =  LibProperty:New(obj, visitor, name .. " expression", config.expression, "", config.unit, errorLevel)
    obj.expr_min = LibProperty:New(obj, visitor, name .. " min", config.min, "", errorLevel, config.unit);
    obj.expr_max = LibProperty:New(obj, visitor, name .. " max", config.max, "", errorLevel, config.unit);
	obj.color = LibProperty:New(obj, visitor, name .. " color", config.color, "", errorLevel, config.unit);
	obj.width = config.width or WidgetHistogram.defaults.width
	obj.height = config.height or WidgetHistogram.defaults.height
	obj.reversed = config.reversed
	obj.char = config.char or WidgetHistogram.defaults.char

	obj.direction = config.direction or WidgetHistogram.defaults.direction --visitor.CFG:FetchRaw(config, "direction", WidgetHistogram.defaults.direction)

	obj.update = config.update or WidgetHistogram.defaults.update -- visitor.CFG:FetchRaw(config, "update", WidgetHistogram.defaults.update)
	
	obj.timer = LibTimer:New("WidgetHistogram.timer " .. obj.widget.name, obj.update, true, obj.Update, obj)

    --QObject::connect(v->GetWrapper(), SIGNAL(_ResizeLCD(int, int, int, int)),
    --    this, SLOT(Resize(int, int, int, int)));

	table.wipe(obj.history)
	
	for i = 1, obj.width do
		tinsert(obj.history, {0, 1, 0, 0})
	end
		
	return obj	
end

-- Delete a StarLibWidgetHistogram object
-- @name StarLibWidgetHistogram.Del
-- @usage StarLibWidgetHistogram:Del([marq])
-- @param marq A StarLibWidgetHistogram object
function WidgetHistogram:Del(hist)
	if not hist then
		hist = self
	end
	pool[hist] = true
	hist.error:Del()
	hist.widget:Del()
	hist.expression:Del()
	hist.expr_min:Del()
	hist.expr_max:Del()
	hist.color:Del()
end

function WidgetHistogram:Resize(rows, cols, old_rows, old_cols)
    local xres = self.visitor.lcd.XRES;
    local yres = self.visitor.lcd.YRES
    local y = rows_ * yres / old_rows;
    local x = cols_ * xres / old_cols;
    local r = row_ * yres / old_rows;
    local c = col_ * xres / old_cols;
    self.rows = round(self.visitor.lcd.LROWS * y / yres);
    self.cols = round(self.visitor.lcd.LCOLS * x / xres);
    self.row = round(self.visitor.lcd.LROWS * r / yres);
    self.col = round(self.visitor.lcd.LCOLS * c / xres);
    self:Update();
end

function WidgetHistogram:Start()
	--self.error:Print("WidgetHistogram:Start")
    if( self.update < 0 or self.active ) then
        return;
	end
    self.timer:Start();
    self:Update();
	self.active = true
end

function WidgetHistogram:Stop()
	--self.error:Print("WidgetHistogram:Stop")
    self.timer:Stop();
	self.active = false
end

function WidgetHistogram:Draw(data)
	self:callback(data)
end

local cache = setmetatable({}, {__mode = "k"})
local function new(...)
	local obj = next(cache)
	if type(obj) == "table" then 
		cache[obj] = nil
	else
		obj = {} 
	end
	for i = 1, select("#", ...) do
		obj[i] = select(i, ...)
	end
	return obj
end

local function del(tbl)
	cache[tbl] = true
end

function WidgetHistogram:Update()
    self.expression:Eval();
    local val = self.expression:P2N();

	assert(type(val) == "number", format("%s: expression is not a number (%s)", self.name, type(val)))
	if type(val) ~= "number" then StarTip:Print("Error", self.name, val); return end
	
	self.color:Eval()
	
	local r, g, b, a
	
	if self.color.is_valid then
		r, g, b, a = self.color.res1, self.color.res2, self.color.res3, self.color.res4
	else
		r, g, b, a = 0, 1, 0
	end
	
	local max, min
    if( self.expr_min.is_valid ) then
        self.expr_min:Eval();
        min = self.expr_min:P2N();
		assert(type(min) == "number", format("%s: expr_min is not a number value (%s)", self.name, type(min)))
    else
        min = self.min;
        if( val < min ) then
            min = val;
		end
    end

    if( self.expr_max.is_valid ) then
        self.expr_max:Eval();
        max = self.expr_max:P2N();
		assert(type(max) == "number", format("%s: expr_max is not a number value (%s)", self.name, type(max)))
    else
        max = self.max;
        if( val > max ) then
            max = val;
		end
    end

    self.min = min;
    self.max = max;

    if( max > min ) then
        self.val = (val - min ) / (max - min);
    else
        self.val = 0.0;
    end

	if self.direction == WidgetHistogram.DIR_EAST then
		tinsert(self.history, 1, new(self.val, r, g, b, a))
		if #self.history > self.width then
			del(self.history[#self.history])
			tremove(self.history, #self.history)
		end
	else		
		tinsert(self.history, new(self.val, r, g, b, a))
		if #self.history > self.width then
			del(self.history[1])
			tremove(self.history, 1)
		end
	end
	
    self:Draw(self.data);
end

function WidgetHistogram:GetOptions(db, callback, data)
	local defaults = WidgetHistogram.defaults
	local options = {
		enabled = {
			name = "Enabled",
			desc = "Whether the histogram's enabled or not",
			type = "toggle",
			get = function() return db.enabled end,
			set = function(info, v) 
				db.enabled = v; 
				db["enabledDirty"] = true 
				if type(callback) == "function" then
					callback(data)
				end
			end,
			order = 1
		},
		length = {
			name = "Histogram length",
			desc = "Enter the histogram's length",
			type = "input",
			pattern = "%d",
			get = function() return tostring(db.length or defaults.length) end,
			set = function(info, v) 
				db.length = tonumber(v); 
				db["lengthDirty"] = true 
				if type(callback) == "function" then
					callback(data)
				end
			end,
			order = 2
		},
		height = {
			name = "Histogram height",
			desc = "Enter the histogram's height",
			type = "input",
			pattern = "%d",
			get = function() return tostring(db.height or defaults.height) end,
			set = function(info, v) 
				db.height = tonumber(v); 
				db["heightDirty"] = true 
				if type(callback) == "function" then
					callback(data)
				end
			end,
			order = 3
		},
		update = {
			name = "Histogram update rate",
			desc = "Enter the histogram's refresh rate",
			type = "input",
			pattern = "%d",
			get = function() return tostring(db.update or defaults.update) end,
			set = function(info, v) 
				db.update = tonumber(v); 
				db["updateDirty"] = true 
				if type(callback) == "function" then
					callback(data)
				end
			end,
			order = 4
		},
		direction = {
			name = "Histogram direction",
			type = "select",
			values = WidgetHistogram.directionList,
			get = function() return db.direction or defaults.direction end,
			set = function(info, v) 
				db.direction = v; 
				db["directionDirty"] = true 
				if type(callback) == "function" then
					callback(data)
				end
			end,
			order = 5
		},
		shownAlways = {
			name = "Shown Always",
			desc = "Whether the frame should be shown always or not",
			type = "toggle",
			get = function() return db.shownAlways end,
			set = function(info, val) 
				db.shownAlways = val 
				db["shownAlwaysDirty"] = true
				if type(callback) == "function" then
					callback(data)
				end
			end,
			order = 6
		},		
		expression = {
			name = "Histogram expression",
			desc = "Enter the histogram's first expression",
			type = "input",
			multiline = true,
			width = "full",
			get = function() return db.expression end,
			set = function(info, v) 
				db.expression = v; 
				db["expressionDirty"] = true 
				if type(callback) == "function" then
					callback(data)
				end
			end,
			order = 7
		},
		min = {
			name = "Histogram min expression",
			desc = "Enter the histogram's minimum expression",
			type = "input",
			multiline = true,
			width = "full",
			get = function() return db.min end,
			set = function(info, v) 
				db.min = v; 
				db["minDirty"] = true 
				if type(callback) == "function" then
					callback(data)
				end
			end,
			order = 8
		
		},
		max = {
			name = "Histogram max expression",
			desc = "Enter the histogram's maximum expression",
			type = "input",
			multiline = true,
			width = "full",
			get = function() return db.max end,
			set = function(info, v) 
				db.max = v; 
				db["maxDirty"] = true 
				if type(callback) == "function" then
					callback(data)
				end
			end,
			order = 9
		},
		color = {
			name = "Histogram color expression",
			desc = "Enter the histogarm's color script",
			type = "input",
			multiline = true,
			width = "full",
			get = function() return db.color end,
			set = function(info, v) 
				db.color = v; 
				db["colorDirty"] = true 
				if type(callback) == "function" then
					callback(data)
				end
			end,
			order = 10
		}
	}
	options.widget = LibWidget:GetOptions(db, callback, data)
	return options
end
